#include <stdio.h>
#include <string.h>
#include "mem.h"
#include "bmp.h"

int SaveBmp(char *filename, IMAGE *PImage)
{
    struct TFileHeader {
        unsigned short bfType;
        long int       bfSize, bfRes1, bfOffBits; } fh;
    struct TImageHeader {
        long int       biSize, biWidth, biHeight;
        unsigned short biPlanes, biBitCount;
        long int       biCompression, biSizeImage;
        long int       biXPelsPerMeter, biYPelsPerMeter;
        long int       biClrUsed, biClrImportant; } ih;
    char               bPal[256][4];
    long int           SuffLen;
    long int           Dummy = 0;
    char             * PBuf;
    short              i;
    FILE             * f;

    if(!(f = fopen(filename, "wb"))) return 1;
    fh.bfType=0x4D42;
    if (PImage->BitPerPixel==8)
    {
        SuffLen=((PImage->Xres+3)/4)*4-(PImage->Xres);
        ih.biSize=0x28;
        ih.biWidth=PImage->Xres;
        ih.biHeight=PImage->Yres;
        ih.biPlanes=1;
        ih.biBitCount=8;
        ih.biCompression=0;
        ih.biSizeImage=(PImage->Xres+SuffLen)*(PImage->Yres);
        ih.biXPelsPerMeter=ih.biYPelsPerMeter=0x2E23; // 300dpi (pixels per meter)
        ih.biClrUsed=ih.biClrImportant=0;
        fh.bfSize=(ih.biSizeImage)+0x0436;
        fh.bfRes1=0;
        fh.bfOffBits=0x0436;
        if (fwrite(&fh, 14, 1, f) != 1) return 1;
        if (fwrite(&ih, 40, 1, f) != 1) return 1;
        for (i=0; i<256; i++)
        {
            bPal[i][0]=PImage->Pal[i].b;
            bPal[i][1]=PImage->Pal[i].g;
            bPal[i][2]=PImage->Pal[i].r;
            bPal[i][3]=0;
        }
        if (fwrite(&bPal, 1024, 1, f) != 1) return 1;
        PBuf=(char *)(PImage->PBitmap.p.p0);
        PBuf+=PImage->NumPixel;
        for (i=0; i<PImage->Yres; i++)
        {
            PBuf-=PImage->Xres;
            if (fwrite(PBuf, PImage->Xres, 1, f) != 1) return 1;
            if (SuffLen>0) { if (fwrite((char *)Dummy, SuffLen, 1, f) != 1) return 1; }
        }
    }
    if (PImage->BitPerPixel==24)
    {
        SuffLen=((PImage->Xres*3+3)/4)*4-(PImage->Xres*3);
        ih.biSize=0x28;
        ih.biWidth=PImage->Xres;
        ih.biHeight=PImage->Yres;
        ih.biPlanes=1;
        ih.biBitCount=24;
        ih.biCompression=0;
        ih.biSizeImage=(PImage->Xres*3+SuffLen)*(PImage->Yres);
        ih.biXPelsPerMeter=ih.biYPelsPerMeter=0x2E23; // 300dpi (pixels per meter)
        ih.biClrUsed=ih.biClrImportant=0;
        fh.bfSize=(ih.biSizeImage)+0x0036;
        fh.bfRes1=0;
        fh.bfOffBits=0x0036;
        if (fwrite(&fh, 14, 1, f) != 1) return 1;
        if (fwrite(&ih, 40, 1, f) != 1) return 1;
        PBuf=(char *)(PImage->PBitmap.p.p0);
        PBuf+=(PImage->NumPixel)*3;
        for (i=0; i<PImage->Yres; i++)
        {
            PBuf-=PImage->Xres*3;
            if (fwrite(PBuf, PImage->Xres*3, 1, f) != 1) return 1;
            if (SuffLen>0) { if (fwrite((char *)Dummy, SuffLen, 1, f) != 1) return 1; }
        }
    }
    fclose(f);
    return 0;
}

int LoadBmp(char *filename, IMAGE *PImage)
{
    struct TFileHeader {
        unsigned short bfType;
        long int       bfSize, bfRes1, bfOffBits; } fh;
    struct TImageHeader {
        long int       biSize, biWidth, biHeight;
        unsigned short biPlanes, biBitCount;
        long int       biCompression, biSizeImage;
        long int       biXPelsPerMeter, biYPelsPerMeter;
        long int       biClrUsed, biClrImportant; } ih;
    char               bPal[256][4];
    long int           SuffLen;
    long int           Dummy = 0;
    char             * PBuf;
    short              i;
    FILE             * f;

    if (!(f = fopen(filename, "rb"))) return 1;
    if (fread(&fh, 14, 1, f) != 1) return 2;
    if (fread(&ih, 40, 1, f) != 1) return 3;
    if (fh.bfType != 0x4D42) return 4;
    PImage->Xres=ih.biWidth;
    PImage->Yres=ih.biHeight;
    PImage->BitPerPixel=ih.biBitCount;
    PImage->NumPixel=(PImage->Xres)*(PImage->Yres);
    if ((PImage->BitPerPixel != 8) && (PImage->BitPerPixel != 24)) return 5;
    if (PImage->BitPerPixel==8)
    {
        SuffLen=((PImage->Xres+3)/4)*4-(PImage->Xres);
        if (fread(&bPal, 1024, 1, f) != 1) return 6;
        for (i=0; i<256; i++)
        {
            PImage->Pal[i].b=bPal[i][0];
            PImage->Pal[i].g=bPal[i][1];
            PImage->Pal[i].r=bPal[i][2];
        }
        GetMemForImage(PImage);
        if (PImage->PBitmap.p.p0==NULL) return 7;
        PBuf=PImage->PBitmap.p.p0;
        PBuf+=PImage->NumPixel;
        for (i=0; i<PImage->Yres; i++)
        {
            PBuf-=PImage->Xres;
            if (fread(PBuf, PImage->Xres, 1, f) != 1) return 8;
            if (SuffLen>0) { if (fread((char *)Dummy, SuffLen, 1, f) != 1) return 8; }
        }
    }
    if (PImage->BitPerPixel==24)
    {
        SuffLen=((PImage->Xres*3+3)/4)*4-(PImage->Xres*3);
        GetMemForImage(PImage);
        if (PImage->PBitmap.p.p0==NULL) return 7;
        PBuf=PImage->PBitmap.p.p0;
        PBuf+=PImage->NumPixel*3;
        for (i=0; i<PImage->Yres; i++)
        {
            PBuf-=PImage->Xres*3;
            if (fread(PBuf, PImage->Xres*3, 1, f) != 1) return 8;
            if (SuffLen>0) { if (fread((char *)Dummy, SuffLen, 1, f) != 1) return 8; }
        }
    }
    fclose(f);
    return 0;
}
