/*********************************************************
 *      HIEND-3D.C - Shows 3D texture in 640x480x256.    *
 * ---> advanced set of 254 interpolating functions <--- *
 *                                                       *
 *   (c) 1997-1999 Stepin Maxim       (stepin@mail.ru)   *
 *********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <dos.h>
#include <i86.h>
#include <string.h>
#include "maindefs.h"
#include "sinus.h"
#include "mem.h"
#include "bmp.h"
#include "gif.h"

#define BUF_X_RES       640
#define BUF_Y_RES       480
#define BUF_SIZE        640*480
#define Z_SCREEN        0x2000
#define Z_MIN           0x1400
#define Z_MAX           0x40000

static struct rmregs {
    long  EDI, ESI, EBP, Res, EBX, EDX, ECX, EAX;
    short flags, ES, DS, FS, GS, IP, CS, SP, SS; } RmRegs;

struct TPoint { TLong X, Y, Z; };

static TLong       Xs={0};
static TLong       Ys={0};
static TLong       Zs={0};
static TLong       Zp={0};
static short       QFlag=3;
static short       MFlag=0;
static short int   ScrPal[256][3];
static int         Step, NumPages, CurPage;
static TLong       Point[8][3];
static int         Face[6][4]={
  {0,1,2,3},{1,5,6,2},{5,4,7,6},{4,0,3,7},{3,2,6,7},{4,5,1,0}};
static TLong       Matrix[3][3];

struct vesainfo {
    char                 Signature[4];
    unsigned char        VersionLo;
    unsigned char        VersionHi;
    unsigned short       PInfoOEMLo;
    unsigned short       PInfoOEMHi;
    unsigned long        Capabilities;
    unsigned short       PModeTableLo;
    unsigned short       PModeTableHi;
    unsigned short       TotalMemory;             // VESA 1.2
    unsigned short       RevVersion;              // VBE 2.0
    unsigned short       POEMNameLo;              // VBE 2.0
    unsigned short       POEMNameHi;              // VBE 2.0
    unsigned short       POEMProductNameLo;       // VBE 2.0
    unsigned short       POEMProductNameHi;       // VBE 2.0
    unsigned short       POEMRevVersionLo;        // VBE 2.0
    unsigned short       POEMRevVersionHi;        // VBE 2.0
    char                 Reserved[222];
    char                 OEMData[256];            // VBE 2.0
};

struct vesamodeinfo {
    unsigned short       ModeAttr;
    unsigned char        WinAAttr;
    unsigned char        WinBAttr;
    unsigned short       WinGranularity;
    unsigned short       WinSize;
    unsigned short       WinASegment;
    unsigned short       WinBSegment;
    unsigned short       WinFuncPtrLo;
    unsigned short       WinFuncPtrHi;
    unsigned short       BytesPerScanLine;
    unsigned short       XResolution;             // VESA 1.2
    unsigned short       YResolution;             // VESA 1.2
    unsigned char        XCharSize;               // VESA 1.2
    unsigned char        YCharSize;               // VESA 1.2
    unsigned char        NumPlanes;               // VESA 1.2
    unsigned char        BitsPerPixel;            // VESA 1.2
    unsigned char        NimBanks;                // VESA 1.2
    unsigned char        MemoryModel;             // VESA 1.2
    unsigned char        BankSize;                // VESA 1.2
    unsigned char        NumPages;                // VESA 1.2
    unsigned char        Reserved1;               // VESA 1.2
    unsigned char        RedSize;                 // VESA 1.2
    unsigned char        RedField;                // VESA 1.2
    unsigned char        GreenSize;               // VESA 1.2
    unsigned char        GreenField;              // VESA 1.2
    unsigned char        BlueSize;                // VESA 1.2
    unsigned char        BlueField;               // VESA 1.2
    unsigned char        ResSize;                 // VESA 1.2
    unsigned char        ResField;                // VESA 1.2
    unsigned char        DACInfo;                 // VESA 1.2
    unsigned char        PVideoBufferLo;          // VBE 2.0
    unsigned char        PVideoBufferHi;          // VBE 2.0
    unsigned char        BackScrOffsetLo;         // VBE 2.0
    unsigned char        BackScrOffsetHi;         // VBE 2.0
    unsigned short       BackScrSize;             // VBE 2.0
    unsigned char        Reserved2[206];          // VBE 2.0
};

extern void NormalLine(int Count, void *PTexture, void *PBuffer, int Xt, int Yt, int dXt, int dYt);
extern void SmoothLine(int Count, void *PTexture, void *PBuffer, int Xt, int Yt, int dXt, int dYt);
extern void PerfectLine(int Count, void *PTexture, void *PBuffer, int Xt, int Yt, int dXt, int dYt);

int InitGraph(short Mode, int *Step, int *NumPages)
{
    union REGS regs;
    struct SREGS sregs;
    unsigned short selector, segment;
    struct vesainfo far *PVesaInfo;
    struct vesamodeinfo far *PVesaModeInfo;

    memset(&regs, 0, sizeof(regs));
    regs.w.ax=0x0100;
    regs.w.bx=0x0020;
    int386(0x31, &regs, &regs);
    segment=regs.w.ax;
    selector=regs.w.dx;

    memset(&RmRegs, 0, sizeof(RmRegs));
    RmRegs.EAX=0x00004F00;
    RmRegs.ES=segment;
    RmRegs.EDI=0;
    segread(&sregs);
    memset(&regs, 0, sizeof(regs));
    regs.w.ax=0x0300;
    regs.h.bl=0x10;
    regs.h.bh=0;
    regs.w.cx=0;
    sregs.es=FP_SEG(&RmRegs);
    regs.x.edi=FP_OFF(&RmRegs);
    int386x(0x31, &regs, &regs, &sregs);
    if ((RmRegs.EAX & 0xFFFF) != 0x004F) return 1;
    PVesaInfo=(struct vesainfo *)(segment<<4);
    printf(" %4.4s\n", PVesaInfo->Signature);
    printf(" %i.%i is detected.\n", PVesaInfo->VersionHi, PVesaInfo->VersionLo);
    printf(" %i Kb video memory.\n", PVesaInfo->TotalMemory*64);
    if (PVesaInfo->TotalMemory<5) return 1;
    if (PVesaInfo->TotalMemory<10) *NumPages=1; else *NumPages=2;
    memset(&RmRegs, 0, sizeof(RmRegs));
    RmRegs.EAX=0x00004F01;
    RmRegs.ECX=Mode;
    RmRegs.ES=segment;
    RmRegs.EDI=0;
    segread(&sregs);
    memset(&regs, 0, sizeof(regs));
    regs.w.ax=0x0300;
    regs.h.bl=0x10;
    regs.h.bh=0;
    regs.w.cx=0;
    sregs.es=FP_SEG(&RmRegs);
    regs.x.edi=FP_OFF(&RmRegs);
    int386x(0x31, &regs, &regs, &sregs);
    if ((RmRegs.EAX & 0xFFFF) != 0x004F) return 2;
    PVesaModeInfo=(struct vesamodeinfo *)(segment<<4);
    if ((PVesaModeInfo->ModeAttr) & 0x01 == 0) return 3;
    if ((PVesaModeInfo->WinSize) != 64) return 4;
    *Step=(PVesaModeInfo->WinSize)/(PVesaModeInfo->WinGranularity);

    memset(&RmRegs, 0, sizeof(RmRegs));
    RmRegs.EAX=0x00004F02;
    RmRegs.EBX=Mode;
    segread(&sregs);
    memset(&regs, 0, sizeof(regs));
    regs.w.ax=0x0300;
    regs.h.bl=0x10;
    regs.h.bh=0;
    regs.w.cx=0;
    sregs.es=FP_SEG(&RmRegs);
    regs.x.edi=FP_OFF(&RmRegs);
    int386x(0x31, &regs, &regs, &sregs);
    if ((RmRegs.EAX & 0xFFFF) != 0x004F) return 5;

    memset(&regs, 0, sizeof(regs));
    regs.w.ax=0x0101;
    regs.w.dx=selector;
    int386(0x31, &regs, &regs);
    return 0;
}

void CloseGraph()
{
    _asm
    {
        mov  ax, 3
        int  10h
    };
}

void SetPalette()
{
    union REGS regs;
    struct SREGS sregs;
    unsigned short selector, segment, i;
    unsigned char *Palette;

    memset(&regs, 0, sizeof(regs));
    regs.w.ax=0x0100;
    regs.w.bx=0x0030;
    int386(0x31, &regs, &regs);
    segment=regs.w.ax;
    selector=regs.w.dx;
    Palette=(unsigned char *)(segment<<4);
    for (i=0; i<256; i++)
    {
        *Palette++=(unsigned char) ScrPal[i][0];
        *Palette++=(unsigned char) ScrPal[i][1];
        *Palette++=(unsigned char) ScrPal[i][2];
    }
    memset(&RmRegs, 0, sizeof(RmRegs));
    RmRegs.EAX=0x00001012;
    RmRegs.EBX=0;
    RmRegs.ECX=256;
    RmRegs.EDX=0;
    RmRegs.ES=segment;
    segread(&sregs);
    memset(&regs, 0, sizeof(regs));
    regs.w.ax=0x0300;
    regs.h.bl=0x10;
    regs.h.bh=0;
    regs.w.cx=0;
    sregs.es=FP_SEG(&RmRegs);
    regs.x.edi=FP_OFF(&RmRegs);
    int386x(0x31, &regs, &regs, &sregs);
    memset(&regs, 0, sizeof(regs));
    regs.w.ax=0x0101;
    regs.w.dx=selector;
    int386(0x31, &regs, &regs);
}

int InitMouse()
{
    union REGS regs;

    memset(&regs, 0, sizeof(regs));
    regs.w.ax=0x0000;
    int386(0x33, &regs, &regs);
    if (regs.w.ax != 0xFFFF) return (regs.w.ax+1);
    memset(&regs, 0, sizeof(regs));
    regs.w.ax=0x000B;
    int386(0x33, &regs, &regs);
    return 0;
}

int MouseButtons()
{
    union REGS regs;

    memset(&regs, 0, sizeof(regs));
    regs.w.ax=0x0003;
    int386(0x33, &regs, &regs);
    return regs.w.bx;
}

void WhereMouse(short int * DeltaX, short int * DeltaY)
{
    union REGS regs;

    memset(&regs, 0, sizeof(regs));
    regs.w.ax=0x000B;
    int386(0x33, &regs, &regs);
    *DeltaX=regs.w.cx;
    *DeltaY=regs.w.dx;
}

int MulDiv(int N1, int N2, int N3);
#pragma aux MulDiv = \
"imul ebx" \
"idiv ecx" \
parm [eax] [ebx] [ecx] \
modify [eax edx] \
value [eax];

int MulDivPlus(int N1, int N2, int N3);
#pragma aux MulDivPlus = \
"imul ebx" \
"idiv ecx" \
"cmp  edx, 0" \
"jle  M1" \
"inc  eax" \
"M1:" \
parm [eax] [ebx] [ecx] \
modify [eax edx] \
value [eax];

int MulDivMinus(int N1, int N2, int N3);
#pragma aux MulDivMinus = \
"imul ebx" \
"idiv ecx" \
"cmp  edx, 0" \
"jge  M1" \
"dec  eax" \
"M1:" \
parm [eax] [ebx] [ecx] \
modify [eax edx] \
value [eax];

void Rotate(TLong *X1, TLong *Y1, long int A)
{
    int Xtmp, Ytmp;

    Xtmp = (*X1).l.l0;
    Ytmp = (*Y1).l.l0;
    if (A)
    _asm
    {
        mov  esi, offset sinus
        mov  edi, A
        mov  eax, Xtmp
        imul dword ptr [esi+edi*4+0x0C00]
        mov  ebx, edx                   // ebx <-- Xs*Cos(F)
        mov  eax, Ytmp
        imul dword ptr [esi+edi*4+0x0C00]
        mov  ecx, edx                   // ecx <-- Ys*Cos(F)
        mov  eax, Ytmp
        imul dword ptr [esi+edi*4+0x0800]
        sub  ebx, edx                   // ebx <-- Xs*Cos(F)-Ys*Sin(F)
        mov  eax, Xtmp
        imul dword ptr [esi+edi*4+0x0800]
        add  ecx, edx                   // ecx <-- Ys*Cos(F)+Xs*Sin(F)
        shl  ebx, 1
        mov  Xtmp, ebx
        shl  ecx, 1
        mov  Ytmp, ecx
    };
    (*X1).l.l0=Xtmp;
    (*Y1).l.l0=Ytmp;
}

void SetVideoBank(int bank)
{
    union REGS regs;
    struct SREGS sregs;

    memset(&RmRegs, 0, sizeof(RmRegs));
    RmRegs.EAX=0x00004F05;
    RmRegs.EBX=0;
    RmRegs.EDX=bank;
    segread(&sregs);
    memset(&regs, 0, sizeof(regs));
    regs.w.ax=0x0300;
    regs.h.bl=0x10;
    regs.h.bh=0;
    regs.w.cx=0;
    sregs.es=FP_SEG(&RmRegs);
    regs.x.edi=FP_OFF(&RmRegs);
    int386x(0x31, &regs, &regs, &sregs);
}

void SetDisplayStart(short int X, short int Y)
{
    union REGS regs;
    struct SREGS sregs;

    memset(&RmRegs, 0, sizeof(RmRegs));
    RmRegs.EAX=0x00004F07;
    RmRegs.EBX=0;
    RmRegs.ECX=X;
    RmRegs.EDX=Y;
    segread(&sregs);
    memset(&regs, 0, sizeof(regs));
    regs.w.ax=0x0300;
    regs.h.bl=0x10;
    regs.h.bh=0;
    regs.w.cx=0;
    sregs.es=FP_SEG(&RmRegs);
    regs.x.edi=FP_OFF(&RmRegs);
    int386x(0x31, &regs, &regs, &sregs);
}

void DrawInBuffer(unsigned char *PBuf, unsigned char *PImage)
{
    int     i, j, k, l, X1, X2, Y1, Y2;
    TLong   Xtl, Ytl, Xtr, Ytr, dXt, dYt;
    struct TPoint       Vert[8];
    struct TPoint       VertP[8];
    short int           VertL, VertR;
    __int64             Ux, Uy, Uz, Vx, Vy, Vz, Nx, Ny, Nz, A, B, Bu, Bv, Bn;
    unsigned char     * PBufTmp;
    int                 count=BUF_SIZE/4;
    extern short int    Xerr, Yerr;
    extern char         Uerr, Werr;
    int                 HALF_X_RES=BUF_X_RES/2;
    int                 HALF_Y_RES=BUF_Y_RES/2;

    srand(239);
    _asm
    {
         cld
         mov  ecx, count
         mov  edi, PBuf
         xor  eax, eax
         rep  stosd
    };
    for (i=0; i<8; i++)
    {
        Vert[i].X.l.l0=Xs.l.l0+\
          MulDiv(Point[i][0].l.l0, Matrix[0][0].l.l0, 0x40000000)+\
          MulDiv(Point[i][1].l.l0, Matrix[0][1].l.l0, 0x40000000)+\
          MulDiv(Point[i][2].l.l0, Matrix[0][2].l.l0, 0x40000000);
        Vert[i].Y.l.l0=Ys.l.l0+\
          MulDiv(Point[i][0].l.l0, Matrix[1][0].l.l0, 0x40000000)+\
          MulDiv(Point[i][1].l.l0, Matrix[1][1].l.l0, 0x40000000)+\
          MulDiv(Point[i][2].l.l0, Matrix[1][2].l.l0, 0x40000000);
        Vert[i].Z.l.l0=Zs.l.l0+\
          MulDiv(Point[i][0].l.l0, Matrix[2][0].l.l0, 0x40000000)+\
          MulDiv(Point[i][1].l.l0, Matrix[2][1].l.l0, 0x40000000)+\
          MulDiv(Point[i][2].l.l0, Matrix[2][2].l.l0, 0x40000000);
    }
    k=Zp.l.l0 >> 4;
    for (i=0; i<8; i++)
    {
        VertP[i].X.l.l0=MulDivPlus(Vert[i].X.l.l0, k, Vert[i].Z.l.l0);
        VertP[i].Y.l.l0=MulDivPlus(Vert[i].Y.l.l0, k, Vert[i].Z.l.l0);
        VertP[i].Z.l.l0=k;
    }
    for (i=0; i<6; i++)
    {
        Nx=(__int64)(Vert[Face[i][1]].Y.l.l0-Vert[Face[i][0]].Y.l.l0) * (__int64)(Vert[Face[i][3]].Z.l.l0-Vert[Face[i][0]].Z.l.l0)-\
           (__int64)(Vert[Face[i][3]].Y.l.l0-Vert[Face[i][0]].Y.l.l0) * (__int64)(Vert[Face[i][1]].Z.l.l0-Vert[Face[i][0]].Z.l.l0);
        Ny=(__int64)(Vert[Face[i][1]].Z.l.l0-Vert[Face[i][0]].Z.l.l0) * (__int64)(Vert[Face[i][3]].X.l.l0-Vert[Face[i][0]].X.l.l0)-\
           (__int64)(Vert[Face[i][3]].Z.l.l0-Vert[Face[i][0]].Z.l.l0) * (__int64)(Vert[Face[i][1]].X.l.l0-Vert[Face[i][0]].X.l.l0);
        Nz=(__int64)(Vert[Face[i][1]].X.l.l0-Vert[Face[i][0]].X.l.l0) * (__int64)(Vert[Face[i][3]].Y.l.l0-Vert[Face[i][0]].Y.l.l0)-\
           (__int64)(Vert[Face[i][3]].X.l.l0-Vert[Face[i][0]].X.l.l0) * (__int64)(Vert[Face[i][1]].Y.l.l0-Vert[Face[i][0]].Y.l.l0);
        if (Nx*Vert[Face[i][0]].X.l.l0+Ny*Vert[Face[i][0]].Y.l.l0+Nz*Vert[Face[i][0]].Z.l.l0<=0) continue;
        Ux=(__int64)(Vert[Face[i][0]].Z.l.l0) * (__int64)(Vert[Face[i][3]].Y.l.l0-Vert[Face[i][0]].Y.l.l0)-\
           (__int64)(Vert[Face[i][0]].Y.l.l0) * (__int64)(Vert[Face[i][3]].Z.l.l0-Vert[Face[i][0]].Z.l.l0);
        Uy=(__int64)(Vert[Face[i][0]].X.l.l0) * (__int64)(Vert[Face[i][3]].Z.l.l0-Vert[Face[i][0]].Z.l.l0)-\
           (__int64)(Vert[Face[i][0]].Z.l.l0) * (__int64)(Vert[Face[i][3]].X.l.l0-Vert[Face[i][0]].X.l.l0);
        Uz=(__int64)(Vert[Face[i][0]].Y.l.l0) * (__int64)(Vert[Face[i][3]].X.l.l0-Vert[Face[i][0]].X.l.l0)-\
           (__int64)(Vert[Face[i][0]].X.l.l0) * (__int64)(Vert[Face[i][3]].Y.l.l0-Vert[Face[i][0]].Y.l.l0);
        Vx=(__int64)(Vert[Face[i][0]].Y.l.l0) * (__int64)(Vert[Face[i][1]].Z.l.l0-Vert[Face[i][0]].Z.l.l0)-\
           (__int64)(Vert[Face[i][0]].Z.l.l0) * (__int64)(Vert[Face[i][1]].Y.l.l0-Vert[Face[i][0]].Y.l.l0);
        Vy=(__int64)(Vert[Face[i][0]].Z.l.l0) * (__int64)(Vert[Face[i][1]].X.l.l0-Vert[Face[i][0]].X.l.l0)-\
           (__int64)(Vert[Face[i][0]].X.l.l0) * (__int64)(Vert[Face[i][1]].Z.l.l0-Vert[Face[i][0]].Z.l.l0);
        Vz=(__int64)(Vert[Face[i][0]].X.l.l0) * (__int64)(Vert[Face[i][1]].Y.l.l0-Vert[Face[i][0]].Y.l.l0)-\
           (__int64)(Vert[Face[i][0]].Y.l.l0) * (__int64)(Vert[Face[i][1]].X.l.l0-Vert[Face[i][0]].X.l.l0);
        Y1=VertP[Face[i][0]].Y.l.l0;
        Y2=VertP[Face[i][0]].Y.l.l0;
        for (j=1; j<4; j++)
        {
            if (VertP[Face[i][j]].Y.l.l0<Y1) Y1=VertP[Face[i][j]].Y.l.l0;
            if (VertP[Face[i][j]].Y.l.l0>Y2) Y2=VertP[Face[i][j]].Y.l.l0;
        }
        if ((Y1>=HALF_Y_RES) || (Y2<-HALF_Y_RES) || (Y1>=Y2)) continue;
        if (Y1<-HALF_Y_RES) Y1=-HALF_Y_RES;
        if (Y2>HALF_Y_RES) Y2=HALF_Y_RES;
        for (j=0; j<4; j++)
        {
            l=(j+1) & 3;
            if ((VertP[Face[i][j]].Y.l.l0<=Y1) && (VertP[Face[i][l]].Y.l.l0>Y1)) VertR=l;
            if ((VertP[Face[i][l]].Y.l.l0<=Y1) && (VertP[Face[i][j]].Y.l.l0>Y1)) VertL=j;
        }
        PBufTmp=PBuf+(Y1+HALF_Y_RES)*BUF_X_RES;
        for (j=Y1; j<Y2; j++)
        {
            while (j>=VertP[Face[i][VertL]].Y.l.l0) VertL=(VertL+3) & 3;
            while (j>=VertP[Face[i][VertR]].Y.l.l0) VertR=(VertR+1) & 3;
            Bu=j*Uy+k*Uz;
            Bv=j*Vy+k*Vz;
            Bn=j*Ny+k*Nz;
            switch (VertR)
            {
                case 0: { A=Ux; B=Bu;       break; }  // U>=0
                case 1: { A=Vx; B=Bv;       break; }  // V>=0
                case 2: { A=Ux-Nx; B=Bu-Bn; break; }  // U<1
                case 3: { A=Vx-Nx; B=Bv-Bn; break; }  // V<1
            }
            if (A>0) { if (B>=0) X2=(-B)/A-1; else X2=(-B-1)/A; }
                else { if (B<=0) X2=(-B)/A-1; else X2=(-B+1)/A; }
            switch (VertL)
            {
                case 0: { A=Vx; B=Bv;       break; }  // V>=0
                case 1: { A=Ux-Nx; B=Bu-Bn; break; }  // U<1
                case 2: { A=Vx-Nx; B=Bv-Bn; break; }  // V<1
                case 3: { A=Ux; B=Bu;       break; }  // U>=0
            }
            if (A>0) { if (B>=0) X1=(-B)/A; else X1=(-B-1)/A+1; }
                else { if (B<=0) X1=(-B)/A; else X1=(-B+1)/A+1; }
            if ((X1<=X2) && (X1<320) && (X2>=-320))
            {
                if (X1<-HALF_X_RES) X1=-HALF_X_RES;
                if (X2>HALF_X_RES-1) X2=HALF_X_RES-1;
                Xtl.l.l0=0x7F0000*(X1*Ux+j*Uy+k*Uz)/(X1*Nx+j*Ny+k*Nz);
                Ytl.l.l0=0x7F0000*(X1*Vx+j*Vy+k*Vz)/(X1*Nx+j*Ny+k*Nz);
                Xtr.l.l0=0x7F0000*(X2*Ux+j*Uy+k*Uz)/(X2*Nx+j*Ny+k*Nz);
                Ytr.l.l0=0x7F0000*(X2*Vx+j*Vy+k*Vz)/(X2*Nx+j*Ny+k*Nz);
                if (X1<X2) dXt.l.l0=(Xtr.l.l0-Xtl.l.l0)/(X2-X1);
                if (X1<X2) dYt.l.l0=(Ytr.l.l0-Ytl.l.l0)/(X2-X1);
                Xerr=rand();
                Yerr=rand();
                Uerr=rand() & 0xFF;
                Werr=rand() & 0xFF;
                switch(QFlag)
                {
                    case 1:
                    {
                        Bu=j*Uy+k*Uz;
                        Bv=j*Vy+k*Vz;
                        Bn=j*Ny+k*Nz;
                        Xtl.l.l0=0x7F0000*(X1*Ux+Bu)/(X1*Nx+Bn);
                        Ytl.l.l0=0x7F0000*(X1*Vx+Bv)/(X1*Nx+Bn);
                        X1+=32;
                        while (X1<=X2)
                        {
                            Xtr.l.l0=0x7F0000*(X1*Ux+Bu)/(X1*Nx+Bn);
                            Ytr.l.l0=0x7F0000*(X1*Vx+Bv)/(X1*Nx+Bn);
                            dXt.l.l0=(Xtr.l.l0-Xtl.l.l0)/32;
                            dYt.l.l0=(Ytr.l.l0-Ytl.l.l0)/32;
                            NormalLine(32, PImage, PBufTmp+X1+HALF_X_RES-32, Xtl.l.l0, Ytl.l.l0, dXt.l.l0, dYt.l.l0);
                            Xtl=Xtr;
                            Ytl=Ytr;
                            X1+=32;
                        }
                        X1-=32;
                        Xtr.l.l0=0x7F0000*(X2*Ux+Bu)/(X2*Nx+Bn);
                        Ytr.l.l0=0x7F0000*(X2*Vx+Bv)/(X2*Nx+Bn);
                        if (X1<X2) dXt.l.l0=(Xtr.l.l0-Xtl.l.l0)/(X2-X1);
                        if (X1<X2) dYt.l.l0=(Ytr.l.l0-Ytl.l.l0)/(X2-X1);
                        NormalLine(X2-X1+1, PImage, PBufTmp+X1+HALF_X_RES, Xtl.l.l0, Ytl.l.l0, dXt.l.l0, dYt.l.l0);
                        break;
                    }
                    case 2:
                    {
                        Bu=j*Uy+k*Uz;
                        Bv=j*Vy+k*Vz;
                        Bn=j*Ny+k*Nz;
                        Xtl.l.l0=0x7F0000*(X1*Ux+Bu)/(X1*Nx+Bn);
                        Ytl.l.l0=0x7F0000*(X1*Vx+Bv)/(X1*Nx+Bn);
                        X1+=32;
                        while (X1<=X2)
                        {
                            Xtr.l.l0=0x7F0000*(X1*Ux+Bu)/(X1*Nx+Bn);
                            Ytr.l.l0=0x7F0000*(X1*Vx+Bv)/(X1*Nx+Bn);
                            dXt.l.l0=(Xtr.l.l0-Xtl.l.l0)/32;
                            dYt.l.l0=(Ytr.l.l0-Ytl.l.l0)/32;
                            SmoothLine(32, PImage, PBufTmp+X1+HALF_X_RES-32, Xtl.l.l0, Ytl.l.l0, dXt.l.l0, dYt.l.l0);
                            Xtl=Xtr;
                            Ytl=Ytr;
                            X1+=32;
                        }
                        X1-=32;
                        Xtr.l.l0=0x7F0000*(X2*Ux+Bu)/(X2*Nx+Bn);
                        Ytr.l.l0=0x7F0000*(X2*Vx+Bv)/(X2*Nx+Bn);
                        if (X1<X2) dXt.l.l0=(Xtr.l.l0-Xtl.l.l0)/(X2-X1);
                        if (X1<X2) dYt.l.l0=(Ytr.l.l0-Ytl.l.l0)/(X2-X1);
                        SmoothLine(X2-X1+1, PImage, PBufTmp+X1+HALF_X_RES, Xtl.l.l0, Ytl.l.l0, dXt.l.l0, dYt.l.l0);
                        break;
                    }
                    case 3:
                    {
                        Bu=j*Uy+k*Uz;
                        Bv=j*Vy+k*Vz;
                        Bn=j*Ny+k*Nz;
                        Xtl.l.l0=0x7F0000*(X1*Ux+Bu)/(X1*Nx+Bn);
                        Ytl.l.l0=0x7F0000*(X1*Vx+Bv)/(X1*Nx+Bn);
                        X1+=32;
                        while (X1<=X2)
                        {
                            Xtr.l.l0=0x7F0000*(X1*Ux+Bu)/(X1*Nx+Bn);
                            Ytr.l.l0=0x7F0000*(X1*Vx+Bv)/(X1*Nx+Bn);
                            dXt.l.l0=(Xtr.l.l0-Xtl.l.l0)/32;
                            dYt.l.l0=(Ytr.l.l0-Ytl.l.l0)/32;
                            PerfectLine(32, PImage, PBufTmp+X1+HALF_X_RES-32, Xtl.l.l0, Ytl.l.l0, dXt.l.l0, dYt.l.l0);
                            Xtl=Xtr;
                            Ytl=Ytr;
                            X1+=32;
                        }
                        X1-=32;
                        Xtr.l.l0=0x7F0000*(X2*Ux+Bu)/(X2*Nx+Bn);
                        Ytr.l.l0=0x7F0000*(X2*Vx+Bv)/(X2*Nx+Bn);
                        if (X1<X2) dXt.l.l0=(Xtr.l.l0-Xtl.l.l0)/(X2-X1);
                        if (X1<X2) dYt.l.l0=(Ytr.l.l0-Ytl.l.l0)/(X2-X1);
                        PerfectLine(X2-X1+1, PImage, PBufTmp+X1+HALF_X_RES, Xtl.l.l0, Ytl.l.l0, dXt.l.l0, dYt.l.l0);
                        break;
                    }
                    case 4:
                    {
                        NormalLine(X2-X1+1, PImage, PBufTmp+X1+HALF_X_RES, Xtl.l.l0, Ytl.l.l0, dXt.l.l0, dYt.l.l0);
                        break;
                    }
                    case 5:
                    {
                        SmoothLine(X2-X1+1, PImage, PBufTmp+X1+HALF_X_RES, Xtl.l.l0, Ytl.l.l0, dXt.l.l0, dYt.l.l0);
                        break;
                    }
                    case 6:
                    {
                        PerfectLine(X2-X1+1, PImage, PBufTmp+X1+HALF_X_RES, Xtl.l.l0, Ytl.l.l0, dXt.l.l0, dYt.l.l0);
                        break;
                    }
                }
            }
            PBufTmp+=BUF_X_RES;
        }
    }
}

void BufferToScreen(unsigned char *PBuf)
{
    int bank;
    int count=BUF_SIZE/4;

    bank=CurPage*5*Step;
    while (count>16384)
    {
        SetVideoBank(bank);
        _asm
        {
             cld
             mov  ecx, 16384
             mov  esi, PBuf
             mov  edi, 0x000A0000
             rep  movsd
        };
        PBuf+=65536;
        bank+=Step;
        count-=16384;
    }
    SetVideoBank(bank);
    _asm
    {
         cld
         mov  ecx, count
         mov  esi, PBuf
         mov  edi, 0x000A0000
         rep  movsd
    };
    if (NumPages==2)
    {
        SetDisplayStart(0, CurPage*512);
        CurPage=1-CurPage;
    }
}

void MagnBuffer(unsigned char *PBuf)  // only for 640x480
{
    TLong PBuf1, PBuf2;

    PBuf1.p.p0=PBuf;
    PBuf2.p.p0=PBuf;
    PBuf1.l.l0+=(210*640+280);
    PBuf2.l.l0+=(BUF_SIZE-80*60);
    _asm
    {
        cld
        mov  ecx, 60
        mov  esi, dword ptr PBuf1
        mov  edi, dword ptr PBuf2
      l01:
        push cx
        mov  cx, 20
        rep  movsd
        pop  cx
        add  esi, 560
        loop l01
    };
    PBuf1.p.p0=PBuf;
    PBuf2.p.p0=PBuf;
    PBuf1.l.l0+=(BUF_SIZE-80*60);
    _asm
    {
        cld
        mov  ecx, 60
        mov  esi, dword ptr PBuf1
        mov  edi, dword ptr PBuf2
      l01:
        push ecx
        mov  ecx, 80
      l02:
        mov  al, [esi]
        mov  ah, al
        shl  eax, 16
        mov  al, [esi]
        mov  ah, al
        mov  [edi], eax
        mov  [edi+4], eax
        mov  [edi+640], eax
        mov  [edi+644], eax
        mov  [edi+1280], eax
        mov  [edi+1284], eax
        mov  [edi+1920], eax
        mov  [edi+1924], eax
        mov  [edi+2560], eax
        mov  [edi+2564], eax
        mov  [edi+3200], eax
        mov  [edi+3204], eax
        mov  [edi+3840], eax
        mov  [edi+3844], eax
        mov  [edi+4480], eax
        mov  [edi+4484], eax
        inc  esi
        add  edi, 8
        loop l02
        pop  ecx
        add  edi, 4480
        loop l01
    };
}

void main(int argc, char * argv[])
{
    short int   Dx, Dy;
    int         Res, i, Mb;
    char        ch;
    IMAGE       Image, Buf;
    char      * filename="TEXT01M.GIF";
    auto char OutStr[30];

    printf("Let's go\n");
    if (argc>1) filename=argv[1];
    Res=LoadGif(filename, &Image);
    if (Res) { printf("LoadGif ERROR %i in file '%s'\n", Res, filename); return; };
    printf("Resolution is %ix%i\n", Image.Xres, Image.Yres);
    if ((Image.Xres!=256) || (Image.Yres!=128))
    { printf("ERROR: Expected resolution is 256x128.\n"); return; };

    Buf.Xres=BUF_X_RES;
    Buf.Yres=BUF_Y_RES;
    Buf.NumPixel=BUF_SIZE;
    Buf.BitPerPixel=8;
    for (i=0; i<256; i++)
    {
        Buf.Pal[i].r=Image.Pal[i].r;
        Buf.Pal[i].g=Image.Pal[i].g;
        Buf.Pal[i].b=Image.Pal[i].b;
    }
    GetMemForImage(&Buf);
    if (Buf.PBitmap.p.p0==NULL) { printf("Buf.GetMem ERROR \n"); return; };

    Res=InitGraph(0x101, &Step, &NumPages);
    if (Res==5) CloseGraph;
    if (Res) { printf("InitGraph ERROR: %i\n", Res); return; }
    CurPage=NumPages-1;
    for (i=0; i<256; i++)
    {
        ScrPal[i][0]=Image.Pal[i].r/4;
        ScrPal[i][1]=Image.Pal[i].g/4;
        ScrPal[i][2]=Image.Pal[i].b/4;
    }
    SetPalette();

    strcpy(OutStr, "That's all folks");
    Res=InitMouse();
    if (Res) { strcpy(OutStr, "Mouse not installed !"); goto Fin; }
    Point[0][0].l.l0=-128*16; Point[0][1].l.l0=-128*16; Point[0][2].l.l0=-128*16;
    Point[1][0].l.l0= 127*16; Point[1][1].l.l0=-128*16; Point[1][2].l.l0=-128*16;
    Point[2][0].l.l0= 127*16; Point[2][1].l.l0= 127*16; Point[2][2].l.l0=-128*16;
    Point[3][0].l.l0=-128*16; Point[3][1].l.l0= 127*16; Point[3][2].l.l0=-128*16;
    Point[4][0].l.l0=-128*16; Point[4][1].l.l0=-128*16; Point[4][2].l.l0= 127*16;
    Point[5][0].l.l0= 127*16; Point[5][1].l.l0=-128*16; Point[5][2].l.l0= 127*16;
    Point[6][0].l.l0= 127*16; Point[6][1].l.l0= 127*16; Point[6][2].l.l0= 127*16;
    Point[7][0].l.l0=-128*16; Point[7][1].l.l0= 127*16; Point[7][2].l.l0= 127*16;
    Zp.l.l0=Z_SCREEN;
    Xs.l.l0=0;
    Ys.l.l0=0;
    Zs.l.l0=Z_SCREEN+128*16;
    Matrix[0][0].l.l0=0x3FFFFFFF; Matrix[0][1].l.l0=0; Matrix[0][2].l.l0=0;
    Matrix[1][0].l.l0=0; Matrix[1][1].l.l0=0x3FFFFFFF; Matrix[1][2].l.l0=0;
    Matrix[2][0].l.l0=0; Matrix[2][1].l.l0=0; Matrix[2][2].l.l0=0x3FFFFFFF;
Start:
    Mb=MouseButtons();
    WhereMouse(&Dx, &Dy);
    if (kbhit())
    {
        ch=getch();
        switch (ch)
        {
            case  27: goto Fin;
            case '-':
            {
                Zs.l.l0+=16;
                if (Zs.l.l0>Z_MAX) Zs.l.l0=Z_MAX;
                break;
            };
            case '+':
            {
                Zs.l.l0-=16;
                if (Zs.l.l0<Z_MIN) Zs.l.l0=Z_MIN;
                break;
            };
            case '/':
            {
                Zs.l.l0=Zs.l.l0*2;
                if (Zs.l.l0>Z_MAX) Zs.l.l0=Z_MAX;
                break;
            };
            case '*':
            {
                Zs.l.l0=Zs.l.l0/2;
                if (Zs.l.l0<Z_MIN) Zs.l.l0=Z_MIN;
                break;
            };
            case ',': { Dx--; break; };
            case '.': { Dx++; break; };
            case 'a': { Dy--; break; };
            case 'A': { Dy--; break; };
            case 'z': { Dy++; break; };
            case 'Z': { Dy++; break; };
            case '1': { QFlag=1; break; };
            case '2': { QFlag=2; break; };
            case '3': { QFlag=3; break; };
            case '4': { QFlag=4; break; };
            case '5': { QFlag=5; break; };
            case '6': { QFlag=6; break; };
            case '7': { QFlag=7; break; };
            case '8': { QFlag=8; break; };
            case '9': { QFlag=9; break; };
            case 'm': { MFlag=!MFlag; break; };
            case 'M': { MFlag=!MFlag; break; };
            case  0:
            {
                ch=getch();
                switch (ch)
                {
                    case 68: { SaveBmp("SCREEN00.BMP", &Buf); break; } //F10
                    case 71:                                           //Home
                    {
                        Xs.l.l0=0;
                        Ys.l.l0=0;
                        Zs.l.l0=Z_SCREEN+128*16;
                        Matrix[0][0].l.l0=0x3FFFFFFF; Matrix[0][1].l.l0=0; Matrix[0][2].l.l0=0;
                        Matrix[1][0].l.l0=0; Matrix[1][1].l.l0=0x3FFFFFFF; Matrix[1][2].l.l0=0;
                        Matrix[2][0].l.l0=0; Matrix[2][1].l.l0=0; Matrix[2][2].l.l0=0x3FFFFFFF;
                        break;
                    };
                    case 72: { Ys.l.l0-=16; break; }    //Up
                    case 75: { Xs.l.l0-=16; break; }    //Left
                    case 77: { Xs.l.l0+=16; break; }    //Right
                    case 80: { Ys.l.l0+=16; break; }    //Down
                }
            }
        }
        ch='^';
    }
    if (Mb & 2)
    {
        Xs.l.l0+=Dx*16;
        Ys.l.l0+=Dy*16;
    } else
    if (Mb & 1)
    {
        Zs.l.l0-=Dy*16;
        if (Zs.l.l0>Z_MAX) Zs.l.l0=Z_MAX;
        if (Zs.l.l0<Z_MIN) Zs.l.l0=Z_MIN;
    } else
    {
        if (Dx>256) Dx-=256;
        if (Dx<-256) Dx+=256;
        Rotate(&Matrix[0][0], &Matrix[2][0], (int)Dx);
        Rotate(&Matrix[0][1], &Matrix[2][1], (int)Dx);
        Rotate(&Matrix[0][2], &Matrix[2][2], (int)Dx);
        if (Dy>256) Dy-=256;
        if (Dy<-256) Dy+=256;
        Rotate(&Matrix[1][0], &Matrix[2][0], (int)Dy);
        Rotate(&Matrix[1][1], &Matrix[2][1], (int)Dy);
        Rotate(&Matrix[1][2], &Matrix[2][2], (int)Dy);
    }
    DrawInBuffer(Buf.PBitmap.p.p0, Image.PBitmap.p.p0);
    if (MFlag) MagnBuffer(Buf.PBitmap.p.p0);
    BufferToScreen(Buf.PBitmap.p.p0);
    goto Start;
Fin:
    CloseGraph();

    FreeMemForImage(&Buf);
    FreeMemForImage(&Image);

    printf("%s\n", OutStr);
//    sound(100);delay(1000);nosound();delay(1000);
}


