;*********************************************************
;*       HIEND-TM.ASM - drawing textured scanline.       *
;* ---> advanced set of 254 interpolating functions <--- *
;*                                                       *
;*   (c) 1997-1999 Stepin Maxim       (stepin@mail.ru)   *
;*********************************************************

        .386P

        include putpixel.inc

        public NormalLine
        public SmoothLine
        public PerfectLine

        Count        equ 8
        PTexture     equ 12
        PBuffer      equ 16
        Xt           equ 20
        Yt           equ 24
        dXt          equ 28
        dYt          equ 32

        .CODE

;****************************************************************************
; Draws a textured scanline with normal algoritm.
; void NormalLine(int Count, void *PTexture, void *PBuffer, int Xt, int Yt, int dXt, int dYt);
;****************************************************************************
NormalLine proc near
        push ebp                        ; preserve caller's stack frame
        mov  ebp, esp                   ; point to our stack frame
        push esi
        push edi

        mov  ecx, [ebp+Count]
        mov  esi, [ebp+PTexture]
        mov  edi, [ebp+PBuffer]
        add  dword ptr [ebp+Xt], 0x00008000
        add  dword ptr [ebp+Yt], 0x00008000
        mov  eax, [ebp+dXt]
        mov  edx, [ebp+dYt]
        xor  ebx, ebx
      L1:
        mov  bx, [ebp+Xt+2]
        mov  bh, [ebp+Yt+2]
        mov  bx, [ebx+esi]
        mov  [edi], bl
        add  [ebp+Xt], eax
        add  [ebp+Yt], edx
        inc  edi
        loop L1

        pop  edi
        pop  esi
        mov  esp, ebp
        pop  ebp                         ; restore caller's stack frame
        ret
NormalLine endp

;****************************************************************************
; Draws a smooth textured scanline with bilinear interpolation.
; void SmoothLine(int Count, void *PTexture, void *PBuffer, int Xt, int Yt, int dXt, int dYt);
;****************************************************************************
SmoothLine proc near
        push ebp                        ; preserve caller's stack frame
        mov  ebp, esp                   ; point to our stack frame
        push esi
        push edi

        mov  ecx, [ebp+dXt]
        sar  ecx, 16
        mov  StepX, cl
        mov  ecx, [ebp+dYt]
        sar  ecx, 16
        mov  StepY, cl
        mov  cx, [ebp+Xt]
        mov  dx, [ebp+Yt]
        mov  esi, [ebp+PTexture]
        mov  edi, [ebp+PBuffer]
        xor  ebx, ebx
        mov  bx, [ebp+Xt+2]
        mov  bh, [ebp+Yt+2]
L2:
        add  Xerr, cx
        jc   M10
        add  Yerr, dx
        jc   M01
        mov  ax, [ebx+esi]
        add  cx, [ebp+dXt]
        adc  bl, StepX
        add  dx, [ebp+dYt]
        adc  bh, StepY
        mov  [edi], al
        inc  edi
        dec  dword ptr [ebp+Count]
        jg   L2
        jmp  L2_end
      M10:
        add  Yerr, dx
        jc   M11
        mov  ax, [ebx+esi+1]
        add  cx, [ebp+dXt]
        adc  bl, StepX
        add  dx, [ebp+dYt]
        adc  bh, StepY
        mov  [edi], al
        inc  edi
        dec  dword ptr [ebp+Count]
        jg   L2
        jmp  L2_end
      M11:
        mov  ax, [ebx+esi+257]
        add  cx, [ebp+dXt]
        adc  bl, StepX
        add  dx, [ebp+dYt]
        adc  bh, StepY
        mov  [edi], al
        inc  edi
        dec  dword ptr [ebp+Count]
        jg   L2
        jmp  L2_end
      M01:
        mov  ax, [ebx+esi+256]
        add  cx, [ebp+dXt]
        adc  bl, StepX
        add  dx, [ebp+dYt]
        adc  bh, StepY
        mov  [edi], al
        inc  edi
        dec  dword ptr [ebp+Count]
        jg   L2
L2_end:
        pop  edi
        pop  esi
        mov  esp, ebp
        pop  ebp                         ; restore caller's stack frame
        ret
SmoothLine endp

;****************************************************************************
; Draws a perfect textured scanline with very tricky algorithm (254 interp. functions).
; void PerfectLine(int Count, void *PTexture, void *PBuffer, int Xt, int Yt, int dXt, int dYt);
;****************************************************************************
PerfectLine proc near
        push ebp                        ; preserve caller's stack frame
        mov  ebp, esp                   ; point to our stack frame
        push esi
        push edi

        mov  ecx, [ebp+dXt]
        sar  ecx, 16
        mov  StepX, cl
        mov  ecx, [ebp+dYt]
        sar  ecx, 16
        mov  StepY, cl
        mov  cx, [ebp+Xt]
        mov  dx, [ebp+Yt]
        mov  esi, [ebp+PTexture]
        mov  edi, [ebp+PBuffer]
        xor  ebx, ebx
        mov  bx, [ebp+Xt+2]
        mov  bh, [ebp+Yt+2]
        xor  eax, eax
L3:
        mov  al, [ebx+esi+0080h]
        jmp  [eax*4+offset TypeOffset]
  T_00:
        add  Xerr, cx
        jc   M_00_1
        add  Yerr, dx
        jc   M_00_3
        PutPixel00
      M_00_1:
        add  Yerr, dx
        jc   M_00_2
        PutPixel01
      M_00_2:
        PutPixel11
      M_00_3:
        PutPixel10
  T_01:
        cmp  dx, 0x4000
        jae  M_01_1
        add  Xerr, cx
        jc   M_01_3
        PutPixel00
      M_01_1:
        add  Xerr, cx
        jc   M_01_2
        PutPixel10
      M_01_2:
        PutPixel11
      M_01_3:
        PutPixel01
  T_02:
        mov  al, ch
        shr  al, 2
        add  al, 0x40
        cmp  dh, al
        ja   M_02_1
        add  Xerr, cx
        jc   M_02_3
        PutPixel00
      M_02_1:
        add  Xerr, cx
        jc   M_02_2
        PutPixel10
      M_02_2:
        PutPixel11
      M_02_3:
        PutPixel01
  T_03:
        mov  al, ch
        shr  al, 1
        add  al, 0x40
        cmp  dh, al
        ja   M_03_1
        add  Xerr, cx
        jc   M_03_3
        PutPixel00
      M_03_1:
        add  Xerr, cx
        jc   M_03_2
        PutPixel10
      M_03_2:
        PutPixel11
      M_03_3:
        PutPixel01
  T_04:
        mov  al, ch
        shr  al, 2
        sub  al, 0x80
        add  al, dh
        jb   M_04_1
        add  Xerr, cx
        jc   M_04_3
        PutPixel00
      M_04_1:
        add  Xerr, cx
        jc   M_04_2
        PutPixel10
      M_04_2:
        PutPixel11
      M_04_3:
        PutPixel01
  T_05:
        or   dx, dx
        js   M_05_1
        add  Xerr, cx
        jc   M_05_3
        PutPixel00
      M_05_1:
        add  Xerr, cx
        jc   M_05_2
        PutPixel10
      M_05_2:
        PutPixel11
      M_05_3:
        PutPixel01
  T_06:
        mov  al, ch
        shr  al, 2
        add  al, 0x80
        cmp  dh, al
        ja   M_06_1
        add  Xerr, cx
        jc   M_06_3
        PutPixel00
      M_06_1:
        add  Xerr, cx
        jc   M_06_2
        PutPixel10
      M_06_2:
        PutPixel11
      M_06_3:
        PutPixel01
  T_07:
        mov  al, ch
        shr  al, 1
        sub  al, 0xC0
        add  al, dh
        jb   M_07_1
        add  Xerr, cx
        jc   M_07_3
        PutPixel00
      M_07_1:
        add  Xerr, cx
        jc   M_07_2
        PutPixel10
      M_07_2:
        PutPixel11
      M_07_3:
        PutPixel01
  T_08:
        mov  al, ch
        shr  al, 2
        sub  al, 0xC0
        add  al, dh
        jb   M_08_1
        add  Xerr, cx
        jc   M_08_3
        PutPixel00
      M_08_1:
        add  Xerr, cx
        jc   M_08_2
        PutPixel10
      M_08_2:
        PutPixel11
      M_08_3:
        PutPixel01
  T_09:
        cmp  dx, 0xC000
        jae  M_09_1
        add  Xerr, cx
        jc   M_09_3
        PutPixel00
      M_09_1:
        add  Xerr, cx
        jc   M_09_2
        PutPixel10
      M_09_2:
        PutPixel11
      M_09_3:
        PutPixel01
;------------------------------8<-----cut------8<-----------------------
;  Sorry for cutting... know-how... :)
;  In this chunk must be blocks with labels T_0A, ... , T_DA.
;------------------------------8<-----cut------8<-----------------------
  T_DB:
        or   cx, cx
        js   M_DB_1
        cmp  dx, 0xC000
        jae  M_DB_3
        PutPixel00
      M_DB_1:
        cmp  dx, 0xC000
        jae  M_DB_2
        PutPixel01
      M_DB_2:
        PutPixel11
      M_DB_3:
        PutPixel10
  T_DC:
        or   cx, cx
        jns  M_DC_1
        or   dx, dx
        jns  M_DC_2
        mov  al, ch
        add  al, dh
        cmp  al, 0x40
        jb   M_DC_3
        PutPixel11
      M_DC_1:
        cmp  dx, 0xC000
        jb   M_DC_3
        PutPixel10
      M_DC_2:
        cmp  cx, 0xC000
        jb   M_DC_3
        PutPixel01
      M_DC_3:
        PutPixel00
  T_DD:
        or   cx, cx
        js   M_DD_1
        cmp  dx, 0xC000
        jae  M_DD_3
        PutPixel00
      M_DD_1:
        or   dx, dx
        js   M_DD_2
        PutPixel01
      M_DD_2:
        PutPixel11
      M_DD_3:
        PutPixel10
  T_DE:
        mov  al, ch
        shl  ax, 1
        add  al, dh
        adc  ah, 0
        shr  ax, 2
        cmp  al, 0x70
        jae  M_DE_1
        cmp  dx, 0xC000
        jae  M_DE_2
        cmp  cx, 0xC000
        jae  M_DE_3
        PutPixel00
      M_DE_1:
        or   cx, cx
        jns  M_DE_2
        cmp  dx, 0x4000
        jb   M_DE_3
        PutPixel11
      M_DE_2:
        PutPixel10
      M_DE_3:
        PutPixel01
  T_DF:
        or   cx, cx
        js   M_DF_1
        cmp  dx, 0xC000
        jae  M_DF_3
        PutPixel00
      M_DF_1:
        cmp  dx, 0x4000
        jae  M_DF_2
        PutPixel01
      M_DF_2:
        PutPixel11
      M_DF_3:
        PutPixel10
  T_E0:
        mov  al, dh
        shl  ax, 1
        add  al, ch
        adc  ah, 0
        shr  ax, 2
        cmp  al, 0x70
        jae  M_E0_1
        cmp  dx, 0xC000
        jae  M_E0_2
        cmp  cx, 0xC000
        jae  M_E0_3
        PutPixel00
      M_E0_1:
        cmp  cx, 0x4000
        jb   M_E0_2
        or   dx, dx
        jns  M_E0_3
        PutPixel11
      M_E0_2:
        PutPixel10
      M_E0_3:
        PutPixel01
  T_E1:
        mov  al, ch
        add  al, dh
        jc   M_E1_1
        or   cx, cx
        js   M_E1_3
        cmp  dx, 0xC000
        jae  M_E1_2
        PutPixel00
      M_E1_1:
        or   dx, dx
        jns  M_E1_3
        cmp  cx, 0x4000
        jb   M_E1_2
        PutPixel11
      M_E1_2:
        PutPixel10
      M_E1_3:
        PutPixel01
  T_E2:
        mov  al, ch
        add  al, dh
        jc   M_E2_1
        cmp  dx, 0xC000
        jae  M_E2_2
        cmp  cx, 0xC000
        jae  M_E2_3
        PutPixel00
      M_E2_1:
        cmp  cx, 0x4000
        jb   M_E2_2
        cmp  dx, 0x4000
        jb   M_E2_3
        PutPixel11
      M_E2_2:
        PutPixel10
      M_E2_3:
        PutPixel01
  T_E3:
        mov  al, ch
        shl  ax, 1
        add  al, dh
        adc  ah, 0
        shr  ax, 2
        cmp  al, 0x50
        jae  M_E3_1
        cmp  dx, 0xC000
        jae  M_E3_2
        or   cx, cx
        js   M_E3_3
        PutPixel00
      M_E3_1:
        cmp  cx, 0x4000
        jb   M_E3_2
        cmp  dx, 0x4000
        jb   M_E3_3
        PutPixel11
      M_E3_2:
        PutPixel10
      M_E3_3:
        PutPixel01
  T_E4:
        cmp  cx, dx
        jae  M_E4_1
        or   dx, dx
        jns  M_E4_2
        cmp  cx, 0xC000
        jae  M_E4_3
        PutPixel10
      M_E4_1:
        or   cx, cx
        jns  M_E4_2
        cmp  dx, 0xC000
        jae  M_E4_3
        PutPixel01
      M_E4_2:
        PutPixel00
      M_E4_3:
        PutPixel11
  T_E5:
        mov  al, dh
        shl  ax, 1
        sub  al, ch
        sbb  ah, -1
        shr  ax, 2
        cmp  al, 0x70
        jb   M_E5_1
        or   dx, dx
        jns  M_E5_2
        cmp  cx, 0xC000
        jae  M_E5_3
        PutPixel10
      M_E5_1:
        cmp  cx, 0x4000
        jb   M_E5_2
        cmp  dx, 0xC000
        jae  M_E5_3
        PutPixel01
      M_E5_2:
        PutPixel00
      M_E5_3:
        PutPixel11
  T_E6:
        cmp  cx, 0xC000
        jae  M_E6_1
        or   dx, dx
        js   M_E6_3
        PutPixel00
      M_E6_1:
        or   dx, dx
        js   M_E6_2
        PutPixel01
      M_E6_2:
        PutPixel11
      M_E6_3:
        PutPixel10
  T_E7:
        or   dx, dx
        js   M_E7_1
        or   cx, cx
        js   M_E7_3
        PutPixel00
      M_E7_1:
        cmp  cx, 0xC000
        jae  M_E7_2
        PutPixel10
      M_E7_2:
        PutPixel11
      M_E7_3:
        PutPixel01
  T_E8:
        or   dx, dx
        js   M_E8_1
        cmp  cx, 0x4000
        jae  M_E8_3
        PutPixel00
      M_E8_1:
        cmp  cx, 0xC000
        jae  M_E8_2
        PutPixel10
      M_E8_2:
        PutPixel11
      M_E8_3:
        PutPixel01
  T_E9:
        or   cx, cx
        js   M_E9_1
        or   dx, dx
        js   M_E9_3
        PutPixel00
      M_E9_1:
        cmp  dx, 0xC000
        jae  M_E9_2
        PutPixel01
      M_E9_2:
        PutPixel11
      M_E9_3:
        PutPixel10
  T_EA:
        or   dx, dx
        jns  M_EA_1
        or   cx, cx
        js   M_EA_2
        mov  al, dh
        sub  al, ch
        cmp  al, 0x40
        jb   M_EA_3
        PutPixel10
      M_EA_1:
        cmp  cx, 0x4000
        jae  M_EA_3
        PutPixel00
      M_EA_2:
        cmp  dx, 0xC000
        jb   M_EA_3
        PutPixel11
      M_EA_3:
        PutPixel01
  T_EB:
        or   dx, dx
        js   M_EB_1
        cmp  cx, 0xC000
        jae  M_EB_3
        PutPixel00
      M_EB_1:
        or   cx, cx
        js   M_EB_2
        PutPixel10
      M_EB_2:
        PutPixel11
      M_EB_3:
        PutPixel01
  T_EC:
        or   cx, cx
        js   M_EC_1
        or   dx, dx
        js   M_EC_3
        PutPixel00
      M_EC_1:
        or   dx, dx
        js   M_EC_2
        PutPixel01
      M_EC_2:
        PutPixel11
      M_EC_3:
        PutPixel10
  T_ED:
        or   dx, dx
        js   M_ED_1
        cmp  cx, 0x4000
        jae  M_ED_3
        PutPixel00
      M_ED_1:
        or   cx, cx
        js   M_ED_2
        PutPixel10
      M_ED_2:
        PutPixel11
      M_ED_3:
        PutPixel01
  T_EE:
        mov  al, ch
        add  al, dh
        jc   M_EE_1
        or   dx, dx
        js   M_EE_2
        cmp  cx, 0xC000
        jae  M_EE_3
        PutPixel00
      M_EE_1:
        or   cx, cx
        jns  M_EE_2
        cmp  dx, 0x4000
        jb   M_EE_3
        PutPixel11
      M_EE_2:
        PutPixel10
      M_EE_3:
        PutPixel01
  T_EF:
        or   cx, cx
        js   M_EF_1
        or   dx, dx
        js   M_EF_3
        PutPixel00
      M_EF_1:
        cmp  dx, 0x4000
        jae  M_EF_2
        PutPixel01
      M_EF_2:
        PutPixel11
      M_EF_3:
        PutPixel10
  T_F0:
        or   dx, dx
        js   M_F0_1
        cmp  cx, 0xC000
        jae  M_F0_3
        PutPixel00
      M_F0_1:
        cmp  cx, 0x4000
        jae  M_F0_2
        PutPixel10
      M_F0_2:
        PutPixel11
      M_F0_3:
        PutPixel01
  T_F1:
        or   dx, dx
        js   M_F1_1
        or   cx, cx
        js   M_F1_3
        PutPixel00
      M_F1_1:
        cmp  cx, 0x4000
        jae  M_F1_2
        PutPixel10
      M_F1_2:
        PutPixel11
      M_F1_3:
        PutPixel01
  T_F2:
        cmp  cx, 0x4000
        jae  M_F2_1
        or   dx, dx
        js   M_F2_3
        PutPixel00
      M_F2_1:
        or   dx, dx
        js   M_F2_2
        PutPixel01
      M_F2_2:
        PutPixel11
      M_F2_3:
        PutPixel10
  T_F3:
        mov  al, dh
        shl  ax, 1
        add  al, ch
        adc  ah, 0
        shr  ax, 2
        cmp  al, 0x50
        jae  M_F3_1
        or   dx, dx
        js   M_F3_2
        cmp  cx, 0xC000
        jae  M_F3_3
        PutPixel00
      M_F3_1:
        cmp  cx, 0x4000
        jb   M_F3_2
        cmp  dx, 0x4000
        jb   M_F3_3
        PutPixel11
      M_F3_2:
        PutPixel10
      M_F3_3:
        PutPixel01
  T_F4:
        or   cx, cx
        js   M_F4_1
        or   dx, dx
        js   M_F4_2
        mov  al, ch
        add  al, dh
        cmp  al, 0xC0
        jae  M_F4_3
        PutPixel00
      M_F4_1:
        cmp  dx, 0x4000
        jae  M_F4_3
        PutPixel01
      M_F4_2:
        cmp  cx, 0x4000
        jae  M_F4_3
        PutPixel10
      M_F4_3:
        PutPixel11
  T_F5:
        mov  al, ch
        shl  ax, 1
        sub  al, dh
        sbb  ah, -1
        shr  ax, 2
        cmp  al, 0x70
        jae  M_F5_1
        cmp  dx, 0x4000
        jb   M_F5_2
        cmp  cx, 0xC000
        jae  M_F5_3
        PutPixel10
      M_F5_1:
        or   cx, cx
        jns  M_F5_2
        cmp  dx, 0xC000
        jae  M_F5_3
        PutPixel01
      M_F5_2:
        PutPixel00
      M_F5_3:
        PutPixel11
  T_F6:
        cmp  cx, dx
        jae  M_F6_1
        cmp  dx, 0x4000
        jb   M_F6_2
        cmp  cx, 0xC000
        jae  M_F6_3
        PutPixel10
      M_F6_1:
        cmp  cx, 0x4000
        jb   M_F6_2
        cmp  dx, 0xC000
        jae  M_F6_3
        PutPixel01
      M_F6_2:
        PutPixel00
      M_F6_3:
        PutPixel11
  T_F7:
        or   cx, cx
        jns  M_F7_1
        or   dx, dx
        js   M_F7_2
        mov  al, ch
        sub  al, dh
        cmp  al, 0x40
        jb   M_F7_3
        PutPixel01
      M_F7_1:
        cmp  dx, 0x4000
        jae  M_F7_3
        PutPixel00
      M_F7_2:
        cmp  cx, 0xC000
        jb   M_F7_3
        PutPixel11
      M_F7_3:
        PutPixel10
  T_F8:
        mov  al, dh
        shl  ax, 1
        sub  al, ch
        sbb  ah, -1
        shr  ax, 2
        cmp  al, 0x50
        jb   M_F8_1
        cmp  dx, 0x4000
        jb   M_F8_2
        cmp  cx, 0xC000
        jae  M_F8_3
        PutPixel10
      M_F8_1:
        cmp  cx, 0x4000
        jb   M_F8_2
        or   dx, dx
        js   M_F8_3
        PutPixel01
      M_F8_2:
        PutPixel00
      M_F8_3:
        PutPixel11
  T_F9:
        or   cx, cx
        js   M_F9_1
        cmp  dx, 0x4000
        jae  M_F9_3
        PutPixel00
      M_F9_1:
        cmp  dx, 0xC000
        jae  M_F9_2
        PutPixel01
      M_F9_2:
        PutPixel11
      M_F9_3:
        PutPixel10
  T_FA:
        mov  al, ch
        shl  ax, 1
        sub  al, dh
        sbb  ah, -1
        shr  ax, 2
        cmp  al, 0x50
        jae  M_FA_1
        cmp  dx, 0x4000
        jb   M_FA_2
        or   cx, cx
        js   M_FA_3
        PutPixel10
      M_FA_1:
        cmp  cx, 0x4000
        jb   M_FA_2
        cmp  dx, 0xC000
        jae  M_FA_3
        PutPixel01
      M_FA_2:
        PutPixel00
      M_FA_3:
        PutPixel11
  T_FB:
        or   cx, cx
        js   M_FB_1
        cmp  dx, 0x4000
        jae  M_FB_3
        PutPixel00
      M_FB_1:
        or   dx, dx
        js   M_FB_2
        PutPixel01
      M_FB_2:
        PutPixel11
      M_FB_3:
        PutPixel10
  T_FC:
        cmp  cx, dx
        jae  M_FC_1
        or   cx, cx
        js   M_FC_3
        cmp  dx, 0x4000
        jb   M_FC_2
        PutPixel10
      M_FC_1:
        or   dx, dx
        js   M_FC_3
        cmp  cx, 0x4000
        jb   M_FC_2
        PutPixel01
      M_FC_2:
        PutPixel00
      M_FC_3:
        PutPixel11
  T_FD:
        or   cx, cx
        js   M_FD_1
        cmp  dx, 0x4000
        jae  M_FD_3
        PutPixel00
      M_FD_1:
        cmp  dx, 0x4000
        jae  M_FD_2
        PutPixel01
      M_FD_2:
        PutPixel11
      M_FD_3:
        PutPixel10

L3_end:
        pop  edi
        pop  esi
        mov  esp, ebp
        pop  ebp                         ; restore caller's stack frame
        ret
PerfectLine endp

        .DATA

public          Xerr, Yerr, Uerr, Werr

StepX           db ?
StepY           db ?
Xerr            dw 0
Yerr            dw 0
Uerr            db 0
Werr            db 0

TypeOffset      label dword
dd offset T_00
dd offset T_01, offset T_02, offset T_03, offset T_04, offset T_05, offset T_06, offset T_07, offset T_08, offset T_09
dd offset T_0A, offset T_0B, offset T_0C, offset T_0D, offset T_0E, offset T_0F, offset T_10, offset T_11, offset T_12
dd offset T_13, offset T_14, offset T_15, offset T_16, offset T_17, offset T_18, offset T_19, offset T_1A, offset T_1B
dd offset T_1C, offset T_1D, offset T_1E, offset T_1F, offset T_20, offset T_21, offset T_22, offset T_23, offset T_24
dd offset T_25, offset T_26, offset T_27, offset T_28, offset T_29, offset T_2A, offset T_2B, offset T_2C, offset T_2D
dd offset T_2E, offset T_2F, offset T_30, offset T_31, offset T_32, offset T_33, offset T_34, offset T_35, offset T_36

dd offset T_37, offset T_38, offset T_39, offset T_3A, offset T_3B, offset T_3C, offset T_3D, offset T_3E, offset T_3F
dd offset T_40, offset T_41, offset T_42, offset T_43, offset T_44, offset T_45, offset T_46, offset T_47, offset T_48
dd offset T_49, offset T_4A, offset T_4B, offset T_4C, offset T_4D, offset T_4E, offset T_4F, offset T_50, offset T_51
dd offset T_52, offset T_53, offset T_54, offset T_55, offset T_56, offset T_57, offset T_58, offset T_59, offset T_5A

dd offset T_5B, offset T_5C, offset T_5D, offset T_5E, offset T_5F, offset T_60, offset T_61, offset T_62, offset T_63
dd offset T_64, offset T_65, offset T_66, offset T_67, offset T_68, offset T_69, offset T_6A, offset T_6B, offset T_6C
dd offset T_6D, offset T_6E, offset T_6F, offset T_70, offset T_71, offset T_72, offset T_73, offset T_74, offset T_75
dd offset T_76, offset T_77, offset T_78, offset T_79, offset T_7A, offset T_7B, offset T_7C, offset T_7D, offset T_7E
dd offset T_7F, offset T_80, offset T_81, offset T_82, offset T_83, offset T_84, offset T_85, offset T_86, offset T_87
dd offset T_88, offset T_89, offset T_8A, offset T_8B, offset T_8C, offset T_8D, offset T_8E, offset T_8F, offset T_90
dd offset T_91, offset T_92, offset T_93, offset T_94, offset T_95, offset T_96, offset T_97, offset T_98, offset T_99
dd offset T_9A

dd offset T_9B, offset T_9C, offset T_9D, offset T_9E, offset T_9F, offset T_A0, offset T_A1, offset T_A2, offset T_A3
dd offset T_A4, offset T_A5, offset T_A6, offset T_A7, offset T_A8, offset T_A9, offset T_AA, offset T_AB, offset T_AC
dd offset T_AD, offset T_AE, offset T_AF, offset T_B0, offset T_B1, offset T_B2, offset T_B3, offset T_B4, offset T_B5
dd offset T_B6, offset T_B7, offset T_B8, offset T_B9, offset T_BA, offset T_BB, offset T_BC, offset T_BD, offset T_BE
dd offset T_BF, offset T_C0, offset T_C1, offset T_C2, offset T_C3, offset T_C4, offset T_C5, offset T_C6, offset T_C7
dd offset T_C8, offset T_C9, offset T_CA, offset T_CB, offset T_CC, offset T_CD, offset T_CE, offset T_CF, offset T_D0
dd offset T_D1, offset T_D2, offset T_D3, offset T_D4, offset T_D5, offset T_D6, offset T_D7, offset T_D8, offset T_D9
dd offset T_DA

dd offset T_DB, offset T_DC, offset T_DD, offset T_DE, offset T_DF, offset T_E0, offset T_E1, offset T_E2, offset T_E3
dd offset T_E4, offset T_E5, offset T_E6, offset T_E7, offset T_E8, offset T_E9, offset T_EA, offset T_EB, offset T_EC
dd offset T_ED, offset T_EE, offset T_EF, offset T_F0, offset T_F1, offset T_F2, offset T_F3, offset T_F4, offset T_F5
dd offset T_F6, offset T_F7, offset T_F8, offset T_F9, offset T_FA, offset T_FB, offset T_FC, offset T_FD, offset T_00
dd offset T_00

LUT1            label byte
db   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0
db 255, 128,  85,  64,  51,  43,  36,  32,  28,  26,  23,  21,  20,  18,  17,   0
db 255, 170, 128, 102,  85,  73,  64,  57,  51,  46,  43,  39,  36,  34,  18,   0
db 255, 191, 153, 128, 109,  96,  85,  77,  70,  64,  59,  55,  51,  36,  20,   0
db 255, 204, 170, 146, 128, 113, 102,  93,  85,  78,  73,  68,  55,  39,  21,   0
db 255, 213, 182, 159, 142, 128, 116, 106,  98,  91,  85,  73,  59,  43,  23,   0
db 255, 219, 191, 170, 153, 139, 128, 118, 109, 102,  91,  78,  64,  46,  26,   0
db 255, 223, 198, 179, 162, 149, 137, 128, 119, 109,  98,  85,  70,  51,  28,   0
db 255, 227, 204, 185, 170, 157, 146, 136, 128, 118, 106,  93,  77,  57,  32,   0
db 255, 230, 209, 191, 177, 164, 153, 146, 137, 128, 116, 102,  85,  64,  36,   0
db 255, 232, 213, 196, 182, 170, 164, 157, 149, 139, 128, 113,  96,  73,  43,   0
db 255, 234, 216, 200, 187, 182, 177, 170, 162, 153, 142, 128, 109,  85,  51,   0
db 255, 235, 219, 204, 200, 196, 191, 185, 179, 170, 159, 146, 128, 102,  64,   0
db 255, 237, 221, 219, 216, 213, 209, 204, 198, 191, 182, 170, 153, 128,  85,   0
db 255, 238, 237, 235, 234, 232, 230, 227, 223, 219, 213, 204, 191, 170, 128,   0
db 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,   0

LUT2            label byte
db   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0
db   0,  17,  18,  20,  21,  23,  26,  28,  32,  36,  43,  51,  64,  85, 128, 255
db   0,  18,  34,  36,  39,  43,  46,  51,  57,  64,  73,  85, 102, 128, 170, 255
db   0,  20,  36,  51,  55,  59,  64,  70,  77,  85,  96, 109, 128, 153, 191, 255
db   0,  21,  39,  55,  68,  73,  78,  85,  93, 102, 113, 128, 146, 170, 204, 255
db   0,  23,  43,  59,  73,  85,  91,  98, 106, 116, 128, 142, 159, 182, 213, 255
db   0,  26,  46,  64,  78,  91, 102, 109, 118, 128, 139, 153, 170, 191, 219, 255
db   0,  28,  51,  70,  85,  98, 109, 119, 128, 137, 149, 162, 179, 198, 223, 255
db   0,  32,  57,  77,  93, 106, 118, 128, 136, 146, 157, 170, 185, 204, 227, 255
db   0,  36,  64,  85, 102, 116, 128, 137, 146, 153, 164, 177, 191, 209, 230, 255
db   0,  43,  73,  96, 113, 128, 139, 149, 157, 164, 170, 182, 196, 213, 232, 255
db   0,  51,  85, 109, 128, 142, 153, 162, 170, 177, 182, 187, 200, 216, 234, 255
db   0,  64, 102, 128, 146, 159, 170, 179, 185, 191, 196, 200, 204, 219, 235, 255
db   0,  85, 128, 153, 170, 182, 191, 198, 204, 209, 213, 216, 219, 221, 237, 255
db   0, 128, 170, 191, 204, 213, 219, 223, 227, 230, 232, 234, 235, 237, 238, 255
db   0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255

end
